@@ -3,11 +3,11 @@ import json
 
 async def test_get_example(jp_fetch):
     # When
-    response = await jp_fetch("jupyterlab-judge", "get_example")
+    response = await jp_fetch("jupyterlab-judge", "get-example")
 
     # Then
     assert response.code == 200
     payload = json.loads(response.body)
     assert payload == {
-        "data": "This is /jupyterlab-judge/get_example endpoint!"
+        "data": "This is /jupyterlab-judge/get-example endpoint!"
     }
\ No newline at end of file
