@@ -1,6 +1,10 @@
 import styled from '@emotion/styled';
 import React, { useState, useContext } from 'react';
-import { JudgeError, JudgePanel } from '../widgets/JudgePanel';
+import {
+  JudgeError,
+  JudgeKernelNotConnectedError,
+  JudgePanel
+} from '../widgets/JudgePanel';
 import { transContext } from '../widgets/JudgeSubmissionArea';
 
 export function SubmissionControl(props: {
@@ -30,7 +34,10 @@ export function SubmissionControl(props: {
               errorDetails
             };
             setInProgress(false);
-            throw e;
+            if (!(e instanceof JudgeKernelNotConnectedError)) {
+              // dont report client network error
+              throw e;
+            }
           }
           setInProgress(false);
         }}
@@ -60,7 +67,8 @@ const ControlButton = styled.button`
   background: var(--jp-brand-color1);
 
   /* Shadow-2 */
-  box-shadow: 0px 0.15px 0.45px rgba(0, 0, 0, 0.11),
+  box-shadow:
+    0px 0.15px 0.45px rgba(0, 0, 0, 0.11),
     0px 0.8px 1.8px rgba(0, 0, 0, 0.13);
   border-radius: 8px;
 
