@@ -11,8 +11,7 @@ export namespace ProblemProvider {
 
   export interface IValidateResult {
     token: string | null; // null if token is not issued. (or not required by problem provider)
-    totalCount: number;
-    acceptedCount: number;
+    results: boolean[];
   }
 
   export type SubmissionStatus = 'WA' | 'TLE' | 'AC' | 'OLE' | 'RE' | 'IE';
@@ -23,8 +22,6 @@ export namespace ProblemProvider {
     userId: string;
     status: SubmissionStatus;
     code: string;
-    cpuTime: number;
-    memory: number;
     acceptedCount: number; // 통과한 TC 수
     totalCount: number; // 전체 TC 수
     createdAt: string;
@@ -36,12 +33,42 @@ export namespace ProblemProvider {
   export interface ISubmissionRequest {
     problemId: string;
     code: string;
-    status: SubmissionStatus;
-    acceptedCount: number; // 통과한 TC 수
-    totalCount: number; // 전체 TC 수
     token: string | null; // to validate status, acceptedCount and totalCount. request null if not required
-    cpuTime: number;
-    memory: number;
     language: 'python';
+    details: (
+      | {
+          status: 'WA';
+          answer: string; // outputs from user's code
+          cpuTime: number;
+          memory: number;
+        }
+      | {
+          status: 'TLE'; // Time Limit Exceed
+          cpuTime: number | null; // null if killed
+          memory: number;
+        }
+      | {
+          status: 'AC';
+          cpuTime: number;
+          memory: number;
+        }
+      | {
+          status: 'OLE';
+          cpuTime: number;
+          memory: number;
+        }
+      | {
+          status: 'RE';
+          cpuTime: number;
+          memory: number;
+          errorName: string;
+          errorValue: string;
+        }
+      | {
+          status: 'IE';
+          cpuTime: number;
+          memory: number;
+        }
+    )[];
   }
 }
