@@ -1,13 +1,13 @@
 import {
   Dialog,
   ISessionContext,
+  ISessionContextDialogs,
   SessionContext,
-  sessionContextDialogs,
   showDialog
 } from '@jupyterlab/apputils';
 import { Message } from '@lumino/messaging';
 import { ITranslator, TranslationBundle } from '@jupyterlab/translation';
-import { textEditorIcon } from '@jupyterlab/ui-components';
+import { LabIcon } from '@jupyterlab/ui-components';
 import {
   CodeEditor,
   CodeEditorWrapper,
@@ -19,16 +19,14 @@ import {
   DocumentWidget,
   IDocumentWidget
 } from '@jupyterlab/docregistry';
-import {
-  IRenderMime,
-  IRenderMimeRegistry,
-  MimeModel
-} from '@jupyterlab/rendermime';
-import { CodeMirrorEditorFactory } from '@jupyterlab/codemirror';
+import { IRenderMimeRegistry } from '@jupyterlab/rendermime';
 import { CommandRegistry } from '@lumino/commands';
 import { OutputArea } from '@jupyterlab/outputarea';
 import { KernelMessage } from '@jupyterlab/services';
-import { IHeader, IStreamMsg } from '@jupyterlab/services/lib/kernel/messages';
+import {
+  IErrorMsg,
+  IStreamMsg
+} from '@jupyterlab/services/lib/kernel/messages';
 import { JudgeModel } from '../model';
 import { ProblemProvider } from '../problemProvider/problemProvider';
 import { ToolbarItems } from '../toolbar';
@@ -39,13 +37,41 @@ import { JudgeTerminal } from './JudgeTerminal';
 import { JudgeSubmissionArea } from './JudgeSubmissionArea';
 import { SubmissionList } from '../components/SubmissionList';
 import { IKernelConnection } from '@jupyterlab/services/lib/kernel/kernel';
+import { JudgeSignal } from '../tokens';
+import { customJudgeColorSvg } from '@team-monolith/cds';
+import { IJudgeProblemPanel, JudgeProblemPanel } from './JudgeProblemPanel';
 
-interface IRunResult {
-  status: 'OK' | 'TLE' | 'OLE' | 'RE';
-  output: string;
-  cpuTime: number;
+function bytesToBase64(bytes: Uint8Array) {
+  const binString = Array.from(bytes, byte => String.fromCodePoint(byte)).join(
+    ''
+  );
+  return btoa(binString);
 }
 
+const JudgeColorLabIcon = new LabIcon({
+  name: 'jupyterlab-judge:problem-icon',
+  svgstr: customJudgeColorSvg
+});
+
+type IRunResult =
+  | {
+      status: 'OK' | 'OLE';
+      output: string;
+      cpuTime: number;
+    }
+  | {
+      status: 'RE';
+      output: string;
+      cpuTime: number;
+      errorValue: string;
+      errorName: string;
+    }
+  | {
+      status: 'TLE';
+      output: string;
+      cpuTime: number | null; // null if killed
+    };
+
 export class JudgeError extends Error {
   constructor(message?: string) {
     super(message);
@@ -53,20 +79,44 @@ export class JudgeError extends Error {
   }
 }
 
+export class JudgeKernelNotConnectedError extends JudgeError {
+  constructor(message?: string) {
+    super(message);
+    this.name = this.constructor.name;
+  }
+}
+
+export class JudgeKernelImplementationError extends JudgeError {
+  constructor(message?: string) {
+    super(message);
+    this.name = this.constructor.name;
+  }
+}
+
+export class JudgeKernelReconnectingFailedError extends JudgeError {
+  constructor(message?: string) {
+    super(message);
+    this.name = this.constructor.name;
+  }
+}
+
+export class ValidationFailedError extends JudgeError {
+  constructor(message?: string) {
+    super(message);
+    this.name = this.constructor.name;
+  }
+}
+
 export namespace JudgePanel {
   export interface IOptions {
-    editorConfig: Partial<CodeEditor.IConfig>;
+    editorServices: IEditorServices;
+    editorConfig: Pick<CodeEditor.IOptions, 'config'>;
     rendermime: IRenderMimeRegistry;
     context: DocumentRegistry.IContext<JudgeModel>;
     translator: ITranslator;
-    submitted: Signal<
-      any,
-      {
-        widget: JudgePanel;
-        problem: ProblemProvider.IProblem;
-        submission: ProblemProvider.ISubmission;
-      }
-    >;
+    sessionContextDialogs: ISessionContextDialogs;
+    submitted: Signal<any, JudgeSignal.ISubmissionArgs>;
+    executed: Signal<any, JudgeSignal.IExecutionArgs>;
 
     judgeSubmissionAreaFactory: (
       options: JudgeSubmissionArea.IOptions
@@ -85,8 +135,11 @@ export class JudgePanel extends BoxPanel {
 
     this._context = options.context;
     this._translator = options.translator;
+    this._sessionContextDialogs = options.sessionContextDialogs;
     this._trans = this._translator.load(TRANSLATOR_DOMAIN);
     this._submitted = options.submitted;
+    this._executed = options.executed;
+    this._rendermime = options.rendermime;
 
     this.id = 'jce-judge-panel';
     this.title.closable = true;
@@ -96,16 +149,17 @@ export class JudgePanel extends BoxPanel {
 
     this._editorWidget = new CodeEditorWrapper({
       model: this.model.codeModel,
-      factory: new CodeMirrorEditorFactory().newInlineEditor,
-      config: { ...options.editorConfig, lineNumbers: true }
+      factory: options.editorServices.factoryService.newInlineEditor,
+      editorOptions: { config: { ...options.editorConfig, lineNumbers: true } }
     });
     this._editorWidget.addClass('jp-JudgePanel-editor');
+    this._editorWidget.node.setAttribute('data-jp-undoer', 'true'); // Activate undo/redo keybindings
+    this._editorWidget.node.setAttribute('data-jp-code-runner', 'true'); // Activate run-code keybinding
 
-    this._markdownRenderer = options.rendermime.createRenderer('text/markdown');
-    this._markdownRenderer.addClass('jp-JudgePanel-markdown');
-    this.renderProblem();
+    const problemPanel = this.createProblemPanel();
+    problemPanel.renderProblem();
     this.model.problemChanged.connect((sender, problem) => {
-      this.renderProblem();
+      problemPanel.renderProblem();
       if (problem?.title) {
         this.title.label = `${problem?.title}.judge`;
       }
@@ -127,7 +181,7 @@ export class JudgePanel extends BoxPanel {
     });
     submissionPanel.addClass('jp-JudgePanel-submissionPanel');
 
-    splitPanel.addWidget(this._markdownRenderer);
+    splitPanel.addWidget(problemPanel);
 
     const rightPanel = new SplitPanel({ orientation: 'vertical', spacing: 0 });
     rightPanel.addClass('jp-JudgePanel-rightPanel');
@@ -160,16 +214,13 @@ export class JudgePanel extends BoxPanel {
     return this._context;
   }
 
-  renderProblem(): void {
-    this._markdownRenderer.renderModel(
-      new MimeModel({
-        data: {
-          // 문제를 불러오지 못했습니다.
-          ['text/markdown']:
-            this.model.problem?.content ??
-            this._trans.__('Problem Not Available.')
-        }
-      })
+  createProblemPanel(): IJudgeProblemPanel {
+    return new JudgeProblemPanel(
+      {
+        model: this.model,
+        translator: this._translator
+      },
+      this._rendermime
     );
   }
 
@@ -234,20 +285,6 @@ export class JudgePanel extends BoxPanel {
   }
 
   public async execute(): Promise<KernelMessage.IExecuteReplyMsg | null> {
-    if (this.session.hasNoKernel) {
-      await sessionContextDialogs.selectKernel(this.session);
-      if (this.session.hasNoKernel) {
-        void showDialog({
-          title: this._trans.__('Cell not executed due to missing kernel'),
-          body: this._trans.__(
-            'The cell has not been executed because no kernel selected. Please select a kernel to execute the cell.'
-          ),
-          buttons: [Dialog.okButton({ label: this._trans.__('Ok') })]
-        });
-        return null;
-      }
-    }
-
     if (this.session.pendingInput) {
       void showDialog({
         title: this._trans.__('Cell not executed due to pending input'),
@@ -259,6 +296,24 @@ export class JudgePanel extends BoxPanel {
       return null;
     }
 
+    if (this.session.hasNoKernel) {
+      const shouldSelect = await this.session.startKernel();
+      if (shouldSelect) {
+        await this._sessionContextDialogs.selectKernel(this.session);
+      }
+    }
+
+    if (this.session.hasNoKernel) {
+      void showDialog({
+        title: this._trans.__('Cell not executed due to missing kernel'),
+        body: this._trans.__(
+          'The cell has not been executed because no kernel selected. Please select a kernel to execute the cell.'
+        ),
+        buttons: [Dialog.okButton({ label: this._trans.__('Ok') })]
+      });
+      return null;
+    }
+
     let reply: KernelMessage.IExecuteReplyMsg | undefined = undefined;
     try {
       const code = this.model.source;
@@ -268,6 +323,16 @@ export class JudgePanel extends BoxPanel {
         this.session,
         {}
       );
+
+      let success = false;
+      if (!reply || reply.content.status === 'ok') {
+        success = true;
+      }
+      this._executed.emit({
+        widget: this,
+        cell: this.model.codeModel,
+        success
+      });
     } catch (e) {
       if (
         e instanceof Error &&
@@ -313,7 +378,7 @@ export class JudgePanel extends BoxPanel {
 
     const oldKernel = this.session.session?.kernel;
     if (!oldKernel) {
-      void sessionContextDialogs.selectKernel(this.session);
+      void this._sessionContextDialogs.selectKernel(this.session);
       this.model.submissionStatus = { type: 'idle' };
       return;
     }
@@ -329,7 +394,7 @@ export class JudgePanel extends BoxPanel {
 
     const kernel = sessionContext.session?.kernel;
     if (!kernel) {
-      void sessionContextDialogs.selectKernel(sessionContext);
+      void this._sessionContextDialogs.selectKernel(sessionContext);
       this.model.submissionStatus = { type: 'idle' };
       return;
     }
@@ -340,42 +405,55 @@ export class JudgePanel extends BoxPanel {
       totalCount: testCases.length
     };
 
-    const waitIdleState = new Promise<void>((resolve, reject) => {
-      const resolveOnIdleState = (
-        sender: ISessionContext,
-        state: KernelMessage.Status
-      ) => {
-        if (state === 'idle') {
-          sessionContext.statusChanged.disconnect(resolveOnIdleState);
-          resolve();
-        }
-      };
+    // Wait until kernelDisplayStatus is idle
+    // Check every second up to 20s
+    // Just uses busy loop, no signal
+    for (let i = 0; i < 20; i++) {
       if (sessionContext.kernelDisplayStatus === 'idle') {
-        resolve();
-      } else {
-        sessionContext.statusChanged.connect(resolveOnIdleState);
+        break;
       }
-    });
+      await new Promise(resolve => setTimeout(resolve, 1000));
+    }
 
-    // Wait up to 20s for the status of the kernel.
-    const KERNEL_TIMEOUT_MS = 20000;
-    let timer: number | undefined;
-    await Promise.race([
-      waitIdleState,
-      new Promise((resolve, reject) => {
-        timer = setTimeout(
-          () =>
-            reject(
-              new JudgeError(
-                this._trans.__('Kernel is not responding. Please try again.')
-              )
-            ),
-          KERNEL_TIMEOUT_MS
+    if (sessionContext.kernelDisplayStatus !== 'idle') {
+      // Many issues have been reported here
+      if (sessionContext.kernelDisplayStatus === 'connecting') {
+        // Network problem:
+        // Allow undefined reconnectAttempt, for custom or unexpected IKernelConnection implementation
+        const reconnectAttempt: number | undefined = (kernel as any)
+          ._reconnectAttempt;
+
+        if (reconnectAttempt === undefined) {
+          // This case must be reported
+          throw new JudgeKernelImplementationError(
+            this._trans.__(
+              'Kernel is still connecting. Please check your network.'
+            )
+          );
+        } else if (reconnectAttempt > 0) {
+          // This have some chance to be a server side problem
+          throw new JudgeKernelReconnectingFailedError(
+            this._trans.__(
+              'Kernel is still connecting. Please check your network.'
+            )
+          );
+        } else if (reconnectAttempt === 0) {
+          // This is likely a client side network problem
+          throw new JudgeKernelNotConnectedError(
+            this._trans.__(
+              'Kernel is still connecting. Please check your network.'
+            )
+          );
+        }
+      } else {
+        // Kernel problem
+        console.warn(
+          `Kernel is still ${sessionContext.kernelDisplayStatus} after 20s`
         );
-      })
-    ]);
-    if (timer) {
-      clearTimeout(timer);
+        throw new JudgeError(
+          this._trans.__('Kernel is not responding. Please try again.')
+        );
+      }
     }
 
     const results: IRunResult[] = [];
@@ -389,38 +467,67 @@ export class JudgePanel extends BoxPanel {
       };
     }
 
-    let status: ProblemProvider.SubmissionStatus | null = null;
     const validateResult = await this.model.validate(
       results.map(result => result.output)
     );
-    if (validateResult.acceptedCount === validateResult.totalCount) {
-      status = 'AC';
-    } else if (results.some(result => result.status === 'RE')) {
-      status = 'RE';
-    } else if (results.some(result => result.status === 'OLE')) {
-      status = 'OLE';
-    } else if (results.some(result => result.status === 'TLE')) {
-      status = 'TLE';
-    } else {
-      status = 'WA';
+
+    if (validateResult === null) {
+      throw new ValidationFailedError(
+        this._trans.__('Validation failed. Please try again')
+      );
     }
 
     await kernel.shutdown();
     kernel.dispose();
 
-    const submission = await this.model.submit({
-      problemId: problem.id,
-      status: status,
-      code,
-      cpuTime:
-        results.map(result => result.cpuTime).reduce((a, b) => a + b, 0) /
-        results.length,
-      acceptedCount: validateResult.acceptedCount,
-      totalCount: validateResult.totalCount,
-      token: validateResult.token,
-      language: 'python',
-      memory: 0
-    });
+    const submission = await this.model.submit(
+      {
+        problemId: problem.id,
+        code,
+        token: validateResult.token,
+        language: 'python',
+        details: results.map((result, index) => {
+          switch (result.status) {
+            case 'OK':
+              if (validateResult.results[index]) {
+                return {
+                  status: 'AC',
+                  cpuTime: result.cpuTime,
+                  memory: 0
+                };
+              } else {
+                return {
+                  status: 'WA',
+                  answer: result.output,
+                  cpuTime: result.cpuTime,
+                  memory: 0
+                };
+              }
+            case 'TLE':
+              return {
+                status: 'TLE',
+                cpuTime: result.cpuTime,
+                memory: 0
+              };
+            case 'OLE':
+              return {
+                status: 'OLE',
+                cpuTime: result.cpuTime,
+                memory: 0
+              };
+            case 'RE':
+              return {
+                status: 'RE',
+                memory: 0,
+                cpuTime: result.cpuTime,
+                errorName: result.errorName,
+                errorValue: result.errorValue
+              };
+          }
+        })
+      },
+      this
+    );
 
     this.model.submissionStatus = { type: 'idle' };
 
@@ -437,37 +544,71 @@ export class JudgePanel extends BoxPanel {
     problem: ProblemProvider.IProblem,
     input: string
   ): Promise<IRunResult> {
+    // Transfer inputs to python kernel first,
+    // and then run the code with timeout.
+
+    // Step 1: Override input function
+    //         Prepare StringIO for input
+    const prepareInput = `
+import io
+import base64
+
+def input(prompt=None):  	
+    r = JUDGE_INPUT_STRING_IO.${
+      problem.inputTransferType === 'one_line' ? 'readline' : 'read'
+    }().rstrip('\\n')
+    if not r:  		
+        return ''
+    return r
+
+JUDGE_INPUT_STRING_IO = io.StringIO()
+`;
+    await kernel.requestExecute(
+      {
+        code: prepareInput,
+        stop_on_error: true
+      },
+      true,
+      {}
+    ).done;
+
+    // Step 2: Push input to StringIO
+    //         Encode input to base64 (to avoid escape problem)
+    //         Divide input to chunks (1MB)
+    const CHUNK_SIZE = 1000000;
+    for (let i = 0; i < input.length; i += CHUNK_SIZE) {
+      const chunk = input.slice(i, i + CHUNK_SIZE);
+      const uint8array = new TextEncoder().encode(chunk);
+      const base64EncodedInput = bytesToBase64(uint8array);
+      const pushInput = `
+JUDGE_INPUT_STRING_IO.write(base64.b64decode('${base64EncodedInput}').decode("utf-8"))
+`;
+      await kernel.requestExecute(
+        {
+          code: pushInput,
+          stop_on_error: true
+        },
+        true,
+        {}
+      ).done;
+    }
+
+    // Step 3: Seek to the beginning of StringIO
+    //         This code is prepended to the user code
+    const seekInput = `
+JUDGE_INPUT_STRING_IO.seek(0)
+`;
+
     const content: KernelMessage.IExecuteRequestMsg['content'] = {
-      code,
+      code: seekInput + code,
       stop_on_error: true,
       allow_stdin: true
     };
 
-    let inputLinesLeft: string[] = [];
-    if (problem.inputTransferType === 'one_line') {
-      inputLinesLeft = input.split(/\r?\n/);
-    } else {
-      inputLinesLeft = [input];
-    }
-
     const startTime = Date.now();
     const future = kernel.requestExecute(content, true, {});
-    future.onStdin = (
-      msg: KernelMessage.IStdinMessage<KernelMessage.StdinMessageType>
-    ) => {
-      if (msg.header.msg_type === 'input_request') {
-        const currentInputLine = inputLinesLeft.shift();
-        future.sendInputReply(
-          {
-            value: currentInputLine ?? '',
-            status: 'ok'
-          },
-          msg.header as IHeader<'input_request'>
-        );
-      }
-    };
 
-    const result: IRunResult = { output: '', status: 'OK', cpuTime: 0 };
+    let result: IRunResult = { output: '', status: 'OK', cpuTime: 0 };
     future.onIOPub = (msg: KernelMessage.IIOPubMessage) => {
       const msgType = msg.header.msg_type;
 
@@ -479,9 +620,17 @@ export class JudgePanel extends BoxPanel {
           }
           break;
         }
-        case 'error':
-          result.status = 'RE';
+        case 'error': {
+          const msgError = msg as IErrorMsg;
+          result = {
+            status: 'RE',
+            errorName: msgError.content.ename,
+            errorValue: msgError.content.evalue,
+            output: result.output,
+            cpuTime: 0
+          };
           break;
+        }
         case 'execute_result':
         case 'display_data':
         case 'clear_output':
@@ -505,17 +654,25 @@ export class JudgePanel extends BoxPanel {
       await kernel.interrupt();
 
       // 강제 종료는 당연히 TLE
-      result.status = 'TLE';
+      // result.status = 'TLE';
+      result = {
+        status: 'TLE',
+        output: result.output,
+        cpuTime: null
+      };
     } else {
       // 강제 종료가 아니더라도 TLE 일 수 있습니다.
       // 우리가 시간을 산정하고, TLE를 부여하는 것은 이 cpuTime 기준입니다.
       // 위에서 setTimeout을 하는 것은 다만 커널을 강제종료하기 위한 기준입니다.
       // cpuTime 언제나 0보다 크고 timelimit 이하인 값입니다.
       const cpuTime = Date.now() - startTime;
+      result.cpuTime = cpuTime;
       if (cpuTime > timelimit) {
-        result.status = 'TLE';
-      } else {
-        result.cpuTime = cpuTime;
+        result = {
+          status: 'TLE',
+          output: result.output,
+          cpuTime: cpuTime
+        };
       }
     }
 
@@ -525,19 +682,14 @@ export class JudgePanel extends BoxPanel {
   private _context: DocumentRegistry.IContext<JudgeModel>;
 
   private _editorWidget: CodeEditorWrapper;
-  private _markdownRenderer: IRenderMime.IRenderer;
+  private _rendermime: IRenderMimeRegistry;
   private _terminal: JudgeTerminal.IJudgeTerminal;
 
-  private _translator: ITranslator;
+  protected _translator: ITranslator;
+  private _sessionContextDialogs: ISessionContextDialogs;
   private _trans: TranslationBundle;
-  private _submitted: Signal<
-    any,
-    {
-      widget: JudgePanel;
-      problem: ProblemProvider.IProblem;
-      submission: ProblemProvider.ISubmission;
-    }
-  >;
+  private _submitted: Signal<any, JudgeSignal.ISubmissionArgs>;
+  private _executed: Signal<any, JudgeSignal.IExecutionArgs>;
 }
 
 export class JudgeDocument extends DocumentWidget<JudgePanel, JudgeModel> {
@@ -569,14 +721,17 @@ export class JudgeDocumentFactory extends ABCWidgetFactory<
    */
   constructor(options: JudgeDocumentFactory.IOptions) {
     super(options.factoryOptions);
+    this._editorServices = options.editorServices;
     this._rendermime = options.rendermime;
     this._commands = options.commands;
     this._editorConfig = options.editorConfig;
+    this._sessionContextDialogs = options.sessionContextDialogs;
     this._judgePanelFactory = options.judgePanelFactory;
     this._judgeSubmissionAreaFactory = options.judgeSubmissionAreaFactory;
     this._judgeTerminalFactory = options.judgeTerminalFactory;
     this._submissionListFactory = options.submissionListFactory;
     this._submitted = options.submitted;
+    this._executed = options.executed;
   }
 
   /**
@@ -587,16 +742,19 @@ export class JudgeDocumentFactory extends ABCWidgetFactory<
   ): JudgeDocument {
     const judgePanel = this._judgePanelFactory({
       rendermime: this._rendermime,
+      editorServices: this._editorServices,
       editorConfig: this._editorConfig,
       context,
       translator: this.translator,
+      sessionContextDialogs: this._sessionContextDialogs,
       submitted: this._submitted,
+      executed: this._executed,
       judgeSubmissionAreaFactory: this._judgeSubmissionAreaFactory,
       judgeTerminalFactory: this._judgeTerminalFactory,
       submissionListFactory: this._submissionListFactory
     });
 
-    judgePanel.title.icon = textEditorIcon;
+    judgePanel.title.icon = JudgeColorLabIcon;
     const widget = new JudgeDocument({
       content: judgePanel,
       context,
@@ -606,9 +764,11 @@ export class JudgeDocumentFactory extends ABCWidgetFactory<
     return widget;
   }
 
+  private _editorServices: IEditorServices;
   private _rendermime: IRenderMimeRegistry;
   private _commands: CommandRegistry;
-  private _editorConfig: Partial<CodeEditor.IConfig>;
+  private _editorConfig: Pick<CodeEditor.IOptions, 'config'>;
+  private _sessionContextDialogs: ISessionContextDialogs;
   private _judgePanelFactory: (options: JudgePanel.IOptions) => JudgePanel;
   private _judgeSubmissionAreaFactory: (
     options: JudgeSubmissionArea.IOptions
@@ -619,14 +779,8 @@ export class JudgeDocumentFactory extends ABCWidgetFactory<
   private _submissionListFactory: (
     options: SubmissionList.IOptions
   ) => JSX.Element;
-  private _submitted: Signal<
-    any,
-    {
-      widget: JudgePanel;
-      problem: ProblemProvider.IProblem;
-      submission: ProblemProvider.ISubmission;
-    }
-  >;
+  private _submitted: Signal<any, JudgeSignal.ISubmissionArgs>;
+  private _executed: Signal<any, JudgeSignal.IExecutionArgs>;
 }
 
 /**
@@ -640,7 +794,8 @@ export namespace JudgeDocumentFactory {
     editorServices: IEditorServices;
     rendermime: IRenderMimeRegistry;
     commands: CommandRegistry;
-    editorConfig: Partial<CodeEditor.IConfig>;
+    editorConfig: Pick<CodeEditor.IOptions, 'config'>;
+    sessionContextDialogs: ISessionContextDialogs;
     /**
      * The factory options associated with the factory.
      */
@@ -656,13 +811,7 @@ export namespace JudgeDocumentFactory {
       options: JudgeTerminal.IOptions
     ) => JudgeTerminal.IJudgeTerminal;
     submissionListFactory: (options: SubmissionList.IOptions) => JSX.Element;
-    submitted: Signal<
-      any,
-      {
-        widget: JudgePanel;
-        problem: ProblemProvider.IProblem;
-        submission: ProblemProvider.ISubmission;
-      }
-    >;
+    submitted: Signal<any, JudgeSignal.ISubmissionArgs>;
+    executed: Signal<any, JudgeSignal.IExecutionArgs>;
   }
 }
